# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import shutil
import uuid
import json
import pytest
import tftest
from unittest.mock import patch, DEFAULT, Mock

pytest_plugins = [
    str("_pytest.pytester"),
]

_LOGGER = logging.getLogger('tftest')

cache_methods = ["setup", "init", "plan", "apply", "output", "destroy"]


@pytest.fixture
def tf(request, fixtures_dir):
  terra = tftest.TerraformTest(
      tfdir='plan_no_resource_changes',
      basedir=fixtures_dir,
      enable_cache=request.param,
  )
  yield terra

  _LOGGER.debug("Removing cache dir")
  try:
    shutil.rmtree(terra.cache_dir)
  except FileNotFoundError:
    _LOGGER.debug("%s does not exists", terra.cache_dir)


@pytest.mark.parametrize("tf", [True], indirect=True)
def test_use_cache(tf):
  """
  Ensures cache is used and runs the execute_command() for first call of the 
  method only
  """
  for method in cache_methods:
    with patch.object(tf, 'execute_command', wraps=tf.execute_command) as mock_execute_command:
      for _ in range(2):
        getattr(tf, method)(use_cache=True)
      assert mock_execute_command.call_count == 1


@pytest.mark.parametrize("tf", [
    pytest.param(
        True,
        id="enable_cache"
    ),
    pytest.param(
        False,
        id="disable_cache"
    ),
], indirect=True)
def test_no_use_cache(tf):
  expected_call_count = 2
  for method in cache_methods:
    with patch.object(tf, 'execute_command', wraps=tf.execute_command) as mock_execute_command:
      for _ in range(expected_call_count):
        getattr(tf, method)(use_cache=False)
      assert mock_execute_command.call_count == expected_call_count


@pytest.mark.parametrize("tf", [True], indirect=True)
def test_use_cache_with_same_tf_var_file(tf, tmp_path):
  """
  Ensures cache is used if the same tf_var_file argument is passed
  within subsequent method calls
  """
  tf_var_file_methods = ["plan", "apply", "destroy"]

  tf_vars_file = tmp_path / (str(uuid.uuid4()) + '.json')
  tf_vars_file.write_text(json.dumps({"foo": "old"}))

  for method in tf_var_file_methods:
    with patch.object(tf, 'execute_command', wraps=tf.execute_command) as mock_execute_command:
      for _ in range(2):
        getattr(tf, method)(use_cache=True, tf_var_file=tf_vars_file)

    assert mock_execute_command.call_count == 1


@pytest.mark.parametrize("tf", [True], indirect=True)
def test_use_cache_with_new_tf_var_file(tf, tmp_path):
  """
  Ensures cache is not used if a different tf_var_file argument is passed
  within subsequent method calls
  """
  tf_var_file_methods = ["plan", "apply", "destroy"]
  expected_call_count = 2

  tf_vars_file = tmp_path / (str(uuid.uuid4()) + '.json')

  for method in tf_var_file_methods:
    tf_vars_file.write_text(json.dumps({"foo": "old"}))
    with patch.object(tf, 'execute_command', wraps=tf.execute_command) as mock_execute_command:
      for _ in range(expected_call_count):
        getattr(tf, method)(use_cache=True, tf_var_file=tf_vars_file)
        tf_vars_file.write_text(json.dumps({"foo": "new"}))

    assert mock_execute_command.call_count == expected_call_count


@pytest.mark.parametrize("tf", [True], indirect=True)
def test_use_cache_with_new_extra_files(tf, tmp_path):
  """
  Ensures cache is not used if a different extra_files argument is passed
  within subsequent method calls
  """
  expected_call_count = 2
  tf_vars_file = tmp_path / (str(uuid.uuid4()) + '.json')
  tf_vars_file.write_text(json.dumps({"foo": "old"}))

  with patch.object(tf, 'execute_command', wraps=tf.execute_command) as mock_execute_command:
    for _ in range(expected_call_count):
      tf.setup(use_cache=True, extra_files=[tf_vars_file])
      tf_vars_file.write_text(json.dumps({"foo": "new"}))

    assert mock_execute_command.call_count == expected_call_count


@pytest.mark.parametrize("tf", [True], indirect=True)
def test_use_cache_with_same_extra_files(tf, tmp_path):
  """
  Ensures cache is used if the same extra_files argument is passed
  within subsequent method calls
  """
  tf_vars_file = tmp_path / (str(uuid.uuid4()) + '.json')
  tf_vars_file.write_text(json.dumps({"foo": "old"}))

  with patch.object(tf, 'execute_command', wraps=tf.execute_command) as mock_execute_command:
    for _ in range(2):
      tf.setup(use_cache=True, extra_files=[tf_vars_file])

    assert mock_execute_command.call_count == 1


@pytest.mark.parametrize("tf", [True], indirect=True)
def test_use_cache_with_new_env(tf):
  """
  Ensures cache is not used if the env attribute is updated
  before subsequent method calls
  """
  expected_call_count = 2
  for method in cache_methods:
    with patch.object(tf, 'execute_command', wraps=tf.execute_command) as mock_execute_command:
      for _ in range(expected_call_count):
        getattr(tf, method)(use_cache=True)
        tf._env["foo"] = "bar"

      assert mock_execute_command.call_count == expected_call_count

      del tf._env["foo"]


@pytest.fixture
def dummy_tf_filepath(tf):
  filepath = os.path.join(tf.tfdir, "bar.txt")
  with open(filepath, "w") as f:
    f.write("old")

  yield filepath

  os.remove(filepath)


@pytest.mark.parametrize("tf", [True], indirect=True)
def test_use_cache_with_new_tf_content(tf, dummy_tf_filepath):
  """
  Ensures cache is not used if the tfdir directory is updated
  before subsequent method calls
  """
  expected_call_count = 2
  for method in cache_methods:
    with patch.object(tf, 'execute_command', wraps=tf.execute_command) as mock_execute_command:
      for _ in range(expected_call_count):
        getattr(tf, method)(use_cache=True)
        with open(dummy_tf_filepath, "w") as f:
          f.write(str(uuid.uuid4()))

      assert mock_execute_command.call_count == expected_call_count



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import tftest

import os


@pytest.fixture
def run_all_apply_out(fixtures_dir):
  tg = tftest.TerragruntTest('tg_apply_all', fixtures_dir, tg_run_all=True)
  tg.setup()
  tg.apply(output=False, tg_non_interactive=True)
  yield tg.output()
  tg.destroy(auto_approve=True, tg_non_interactive=True)


@pytest.fixture
def run_all_plan_output(fixtures_dir):
  tg = tftest.TerragruntTest('tg_apply_all', fixtures_dir, tg_run_all=True)
  tg.setup()
  return tg.plan(output=True, tg_working_dir=os.path.join(fixtures_dir, 'tg_apply_all'))


@pytest.fixture
def plan_foo_output(fixtures_dir):
  tg = tftest.TerragruntTest(os.path.join(
      'tg_apply_all', 'foo'), fixtures_dir)
  tg.setup()
  return tg.plan(output=True)


@pytest.fixture
def bar_output(fixtures_dir):
  tg = tftest.TerragruntTest(os.path.join(
      'tg_apply_all', 'bar'), fixtures_dir)
  tg.setup()
  tg.apply(output=False, tg_non_interactive=True)
  yield tg.output()
  tg.destroy(auto_approve=True, tg_non_interactive=True)


def test_run_all_apply(run_all_apply_out):
  triggers = [o["triggers"] for o in run_all_apply_out]
  assert [{'name': 'foo', 'template': 'sample template foo'}] in triggers
  assert [{'name': 'bar', 'template': 'sample template bar'}] in triggers
  assert [{'name': 'one', 'template': 'sample template one'},
          {'name': 'two', 'template': 'sample template two'}] in triggers
  assert len(run_all_apply_out) == 3


def test_tg_single_directory_apply(bar_output):
  assert bar_output["triggers"] == [
      {'name': 'bar', 'template': 'sample template bar'}]


def test_run_all_plan(run_all_plan_output):
  triggers = [o.outputs["triggers"] for o in run_all_plan_output]
  assert [{'name': 'foo', 'template': 'sample template foo'}] in triggers
  assert [{'name': 'bar', 'template': 'sample template bar'}] in triggers
  assert [{'name': 'one', 'template': 'sample template one'},
          {'name': 'two', 'template': 'sample template two'}] in triggers
  assert len(run_all_plan_output) == 3


def test_tg_single_directory_plan(plan_foo_output):
  assert plan_foo_output.outputs['triggers'] == [
      {'name': 'foo', 'template': 'sample template foo'}]
  assert plan_foo_output.variables['names'] == ['foo']



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test init and apply using an actual example."

import pytest
import tftest


@pytest.fixture
def output(fixtures_dir):
  tf = tftest.TerraformTest('apply', fixtures_dir)
  tf.setup()
  tf.apply()
  yield tf.output()
  tf.destroy(**{"auto_approve": True})


def test_apply(output):
  value = output['triggers']
  assert len(value) == 2
  assert list(value[0].keys()) == ['name', 'template']
  assert value[0]['name'] == 'one'



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test fixture file setup and removal."

import os
import tempfile
import tftest


def test_setup_files():
  with tempfile.TemporaryDirectory() as tmpdir:
    try:
      with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        tf = tftest.TerraformTest(tmpdir)
        tf.setup(extra_files=[tmpfile.name])
        assert os.path.exists(os.path.join(
            tmpdir, os.path.basename(tmpfile.name)))
        tf = None
        assert not os.path.exists(os.path.join(
            tmpdir, os.path.basename(tmpfile.name)))
    finally:
      os.unlink(tmpfile.name)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple Python wrapper for Terraform test fixtures.

See documentation in the TerraformTest class for usage. Terraform wrapping
inspired by https://github.com/beelit94/python-terraform
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections
import glob
import inspect
import itertools
import json
import logging
import os
import pickle
import re
import shutil
import stat
import subprocess
import sys
import tempfile
import weakref

from functools import partial
from hashlib import sha1
from pathlib import Path
from typing import List

__version__ = '1.8.5'

_LOGGER = logging.getLogger('tftest')

TerraformCommandOutput = collections.namedtuple('TerraformCommandOutput',
                                                'retcode out err')

TerraformStateResource = collections.namedtuple(
    'TerraformStateResource', 'key provider type attributes depends_on raw')


class TerraformTestError(Exception):

  @property
  def cmd_error(self):
    return self.args[1] if len(self.args) > 1 else None


_TG_BOOL_ARGS = [
    "no_auto_init",
    "no_auto_retry",
    "source_update",
    "ignore_dependency_errors",
    "ignore_dependency_order",
    "include_external_dependencies",
    "check",
    "debug",
    'non_interactive',
    'ignore_external_dependencies',
]

_TG_KV_ARGS = [
    "iam_role",
    "config",
    "tfpath",
    "working_dir",
    "download_dir",
    "source",
    "exclude_dir",
    "include_dir",
    "hclfmt_file",
]


def parse_args(init_vars=None, tf_vars=None, targets=None, **kw):
  """Convert method arguments for use in Terraform commands.

  Args:
    init_vars: dict of key/values converted to -backend-config='k=v' form, or
      string argument converted to -backend-config=arg
    tf_vars: dict of key/values converted to -var k=v form.
    **kw: converted to the appropriate Terraform flag.

  Returns:
    A list of command arguments for use with subprocess.
  """
  cmd_args = []

  cmd_args += [
      f'--terragrunt-{arg.replace("_", "-")}' for arg in _TG_BOOL_ARGS
      if kw.get(f"tg_{arg}")
  ]
  for arg in _TG_KV_ARGS:
    if kw.get(f"tg_{arg}"):
      cmd_args += [f'--terragrunt-{arg.replace("_", "-")}',
                   kw[f"tg_{arg}"]]
  if kw.get('tg_parallelism'):
    cmd_args.append(f'--terragrunt-parallelism {kw["tg_parallelism"]}')
  if isinstance(kw.get('tg_override_attr'), dict):
    cmd_args += [
        '--terragrunt-override-attr={}={}'.format(k, v)
        for k, v in kw.get('tg_override_attr').items()
    ]

  if kw.get('auto_approve'):
    cmd_args.append('-auto-approve')
  if kw.get('backend') is False:
    cmd_args.append('-backend=false')
  if kw.get('color') is False:
    cmd_args.append('-no-color')
  if kw.get('force_copy'):
    cmd_args.append('-force-copy')
  if kw.get('input') is False:
    cmd_args.append('-input=false')
  if kw.get('json_format') is True:
    cmd_args.append('-json')
  if kw.get('lock') is False:
    cmd_args.append('-lock=false')
  if kw.get('plugin_dir'):
    cmd_args += ['-plugin-dir', kw['plugin_dir']]
  if kw.get('refresh') is False:
    cmd_args.append('-refresh=false')
  if kw.get('state'):
    cmd_args += ['-state', kw['state']]
  if kw.get('upgrade'):
    cmd_args.append('-upgrade')
  if isinstance(init_vars, dict):
    cmd_args += [
        '-backend-config={}={}'.format(k, v) for k, v in init_vars.items()
    ]
  elif isinstance(init_vars, str):
    cmd_args += ['-backend-config', '{}'.format(init_vars)]
  if tf_vars:
    cmd_args += list(
        itertools.chain.from_iterable(
            ("-var",
             "{}={}".format(k,
                            json.dumps(v) if isinstance(v, (dict,
                                                            list)) else v))
            for k, v in tf_vars.items()))
  if targets:
    cmd_args += [("-target={}".format(t)) for t in targets]
  if kw.get('tf_var_file'):
    tf_var_file = kw['tf_var_file']
    if isinstance(tf_var_file, (list, tuple)):
      cmd_args += ['-var-file={}'.format(v) for v in tf_var_file]
    else:
      cmd_args.append('-var-file={}'.format(tf_var_file))
  return cmd_args


class TerraformJSONBase(collections.abc.Mapping):
  "Base class for JSON wrappers."

  def __init__(self, raw):
    self._raw = raw

  def __bytes__(self):
    return bytes(self._raw)

  def __getitem__(self, index):
    return self._raw[index]

  def __iter__(self):
    return iter(self._raw)

  def __len__(self):
    return len(self._raw)

  def __str__(self):
    return str(self._raw)


class TerraformValueDict(TerraformJSONBase):
  "Minimal wrapper to directly expose outputs or variables."

  def __init__(self, raw):
    super(TerraformValueDict, self).__init__(raw)
    # only matters for outputs
    self.sensitive = tuple(k for k, v in raw.items() if v.get('sensitive'))

  def __getattr__(self, name):
    if isinstance(name, str) and name[:2] == name[-2:] == '__':
      # skip non-existing dunder method lookups
      raise AttributeError(name)
    return getattr(self._raw, name)

  def __getitem__(self, name):
    return self._raw[name].get('value')


class TerraformPlanModule(TerraformJSONBase):
  "Minimal wrapper for parsed plan output modules."

  def __init__(self, raw):
    super(TerraformPlanModule, self).__init__(raw)
    prefix = raw.get('address', '')
    self._strip = 0 if not prefix else len(prefix) + 1
    self._modules = self._resources = None

  @property
  def child_modules(self):
    if self._modules is None:
      self._modules = dict(
          (mod['address'][self._strip:], TerraformPlanModule(mod))
          for mod in self._raw.get('child_modules', {}))
    return self._modules

  @property
  def resources(self):
    if self._resources is None:
      self._resources = dict((res['address'][self._strip:], res)
                             for res in self._raw.get('resources', []))
    return self._resources


class TerraformPlanOutput(TerraformJSONBase):
  "Minimal wrapper for Terraform plan JSON output."

  def __init__(self, raw):
    super(TerraformPlanOutput, self).__init__(raw)
    planned_values = raw.get('planned_values', {})
    self.root_module = TerraformPlanModule(planned_values.get(
        'root_module', {}))
    self.outputs = TerraformValueDict(planned_values.get('outputs', {}))
    self.resource_changes = dict(
        (v['address'], v) for v in self._raw.get('resource_changes', {}))
    # there might be no variables defined
    self.variables = TerraformValueDict(raw.get('variables', {}))
    self.prior_root_module = TerraformPlanModule(
        raw.get('prior_state', {}).get('values', {}).get('root_module', {}))

  @property
  def resources(self):
    return self.root_module.resources

  @property
  def modules(self):
    return self.root_module.child_modules

  @property
  def prior_resources(self):
    return self.prior_root_module.resources

  def __getattr__(self, name):
    return self._raw[name]

  def __getstate__(self):
    return self.__dict__

  def __setstate__(self, d):
    self.__dict__.update(d)


class TerraformState(TerraformJSONBase):
  "Minimal wrapper for Terraform state JSON format."

  def __init__(self, raw):
    super(TerraformState, self).__init__(raw)
    self.outputs = TerraformValueDict(raw.get('outputs', {}))
    self._resources = None

  @property
  def resources(self):
    if not self._resources:
      resources = {}
      for res in self._raw['resources']:
        name = '%s.%s.%s' % (res.get('module'), res.get('type'),
                             res.get('name'))
        resources[name] = res
      self._resources = resources
    return self._resources

  def __getattr__(self, name):
    return self._raw[name]

  def __getstate__(self):
    return self.__dict__

  def __setstate__(self, d):
    self.__dict__.update(d)


class TerraformTest(object):
  """Helper class for use in testing Terraform modules.

  This helper class can be used to set up fixtures in Terraform tests, so that
  the usual Terraform commands (init, plan, apply, output, destroy) can be run
  on a module. Configuration is done at instantiation first, by passing in the
  Terraform root module path, and then in the setup method through files that
  will be temporarily linked in the module, and Terraform variables.

  The standard way of using this is by calling setup to configure the module
  through temporarily linked Terraform files and variables, run one or more
  Terraform commands, then check command output, state, or created resources
  from individual tests.

  The local .terraform directory (including local state) and any linked file
  are removed when the instance is garbage collected. Destroy needs to be
  called explicitly using destroy().

  Args:
    tfdir: the Terraform module directory to test, either an absolute path, or
      relative to basedir.
    basedir: optional base directory to use for relative paths, defaults to the
      directory above the one this module lives in.
    binary: path to the Terraform command.
    env: a dict with custom environment variables to pass to terraform.
    enable_cache: Determines if the caching enabled for specific methods
    cache_dir: optional base directory to use for caching, defaults to
      the directory of the python file that instantiates this class
  """

  def __init__(self, tfdir, basedir=None, binary='terraform', env=None,
               enable_cache=False, cache_dir=None):
    """Set Terraform folder to operate on, and optional base directory."""
    self._basedir = basedir or os.getcwd()
    self.binary = binary
    self.tfdir = self._abspath(tfdir)
    self._env = env or {}
    self.env = os.environ.copy()
    self.tg_run_all = False
    self._plan_formatter = lambda out: TerraformPlanOutput(json.loads(out))
    self._output_formatter = lambda out: TerraformValueDict(
        json.loads(out))
    self.enable_cache = enable_cache
    if not cache_dir:
      self.cache_dir = Path(os.path.dirname(
          inspect.stack()[1].filename)) / ".tftest-cache"
    else:
      self.cache_dir = Path(cache_dir)
    if env is not None:
      self.env.update(env)

  @classmethod
  def _cleanup(cls, tfdir, filenames, deep=True, restore_files=False):
    """Remove linked files, .terraform and/or .terragrunt-cache folder at instance deletion."""

    def remove_readonly(func, path, excinfo):
      _LOGGER.warning(f'Issue deleting file {path}, caused by {excinfo}')
      os.chmod(path, stat.S_IWRITE)
      func(path)

    _LOGGER.debug('cleaning up %s %s', tfdir, filenames)
    for filename in filenames:
      path = os.path.join(tfdir, filename)
      os.unlink(path)
    if not deep:
      return
    path = os.path.join(tfdir, '.terraform')
    if os.path.isdir(path):
      shutil.rmtree(path, onerror=remove_readonly)
    path = os.path.join(tfdir, 'terraform.tfstate')
    if os.path.isfile(path):
      os.unlink(path)
    path = os.path.join(tfdir, '**', '.terragrunt-cache*')
    for tg_dir in glob.glob(path, recursive=True):
      if os.path.isdir(tg_dir):
        shutil.rmtree(tg_dir, onerror=remove_readonly)
    _LOGGER.debug(
        'Restoring original TF files after prevent destroy changes')
    if restore_files:
      for bkp_file in Path(tfdir).rglob('*.bkp'):
        try:
          shutil.copy(str(bkp_file),
                      f'{str(bkp_file).strip(".bkp")}')
        except (IOError, OSError):
          _LOGGER.exception(
              f'Unable to restore terraform file {bkp_file.resolve()}')
          raise TerraformTestError(
              f'Restore of terraform file ({bkp_file.resolve()}) failed')
        else:
          bkp_file.unlink(True)

  def _abspath(self, path):
    """Make relative path absolute from base dir."""
    return path if os.path.isabs(path) else os.path.join(self._basedir, path)

  def _dirhash(self, directory, hash, ignore_hidden=False,
               exclude_directories=[], excluded_extensions=[]):
    """Returns hash of directory's file contents"""
    assert Path(directory).is_dir()
    try:
      dir_iter = sorted(Path(directory).iterdir(),
                        key=lambda p: str(p).lower())
    except FileNotFoundError:
      return hash
    for path in dir_iter:
      if path.is_file():
        if ignore_hidden and path.name.startswith("."):
          continue
        if path.suffix in excluded_extensions:
          continue
        with open(path, "rb") as f:
          for chunk in iter(lambda: f.read(4096), b""):
            hash.update(chunk)
      elif path.is_dir() and path.name not in exclude_directories:
        hash = self._dirhash(path, hash, ignore_hidden=ignore_hidden,
                             exclude_directories=exclude_directories,
                             excluded_extensions=excluded_extensions)
    return hash

  def generate_cache_hash(self, method_kwargs):
    """Returns a hash value using the instance's attributes and method keyword arguments"""
    params = {
        **{
            k: v for k, v in self.__dict__.items()
            # only uses instance attributes that are involved in the results of
            # the decorated method
            if k in ["binary", "_basedir", "tfdir", "_env"]
        },
        **method_kwargs,
    }

    # creates hash of file contents
    for path_param in ["extra_files", "tf_var_file"]:
      if path_param in method_kwargs:
        if isinstance(method_kwargs[path_param], list):
          params[path_param] = [
              sha1(open(fp, 'rb').read()).hexdigest()
              for fp in method_kwargs[path_param]
          ]
        else:
          params[path_param] = sha1(
              open(method_kwargs[path_param], 'rb').read()).hexdigest()

    # creates hash of all file content within tfdir
    # excludes .terraform/, hidden files, tfstate files from being used for hash
    params["tfdir"] = self._dirhash(self.tfdir, sha1(), ignore_hidden=True,
                                    exclude_directories=[".terraform"],
                                    excluded_extensions=['.backup', '.tfstate'
                                                         ]).hexdigest()

    return sha1(
        json.dumps(params, sort_keys=True,
                   default=str).encode("cp037")).hexdigest() + ".pickle"

  def _cache(func):

    def cache(self, **kwargs):
      """
      Runs the tftest instance method or retreives the cache value if it exists

      Args:
          kwargs: Keyword argument that are passed to the decorated method
      Returns:
          Output of the tftest instance method
      """
      _LOGGER.info("Cache decorated method: %s", func.__name__)

      if not self.enable_cache:
        return func(self, **kwargs)
      elif not kwargs.get("use_cache", False):
        return func(self, **kwargs)

      cache_dir = self.cache_dir / \
          Path(sha1(self.tfdir.encode("cp037")).hexdigest()) / \
          Path(func.__name__)
      cache_dir.mkdir(parents=True, exist_ok=True)

      hash_filename = self.generate_cache_hash(kwargs)
      cache_key = cache_dir / hash_filename
      _LOGGER.debug("Cache key: %s", cache_key)

      try:
        f = cache_key.open("rb")
      except OSError:
        _LOGGER.debug("Could not read cache path")
      else:
        _LOGGER.info("Getting output from cache")
        return pickle.load(f)

      _LOGGER.info("Running command")
      out = func(self, **kwargs)

      if out:
        # the hash value will now include any changes
        # to the tfdir directory including any terragrunt
        # generated files
        hash_filename = self.generate_cache_hash(kwargs)
        cache_key = cache_dir / hash_filename
        _LOGGER.debug("Cache key: %s", cache_key)

        _LOGGER.info("Writing command to cache")
        try:
          f = cache_key.open("wb")
        except OSError as e:
          _LOGGER.error("Cache could not write path")
        else:
          with f:
            pickle.dump(out, f, pickle.HIGHEST_PROTOCOL)

      return out

    return cache

  @_cache
  def setup(self, extra_files=None, plugin_dir=None, init_vars=None,
            backend=True, cleanup_on_exit=True, disable_prevent_destroy=False,
            workspace_name=None, use_cache=False, **kw):
    """Setup method to use in test fixtures.

    This method prepares a new Terraform environment for testing the module
    specified at init time, and returns init output.

    Args:
      extra_files: list of absolute or relative to base paths to be linked in
        the root module folder
      plugin_dir: path to a plugin directory to be used for Terraform init, eg
        built with terraform-bundle
      init_vars: Terraform backend configuration variables
      backend: Terraform backend argument
      cleanup_on_exit: remove .terraform and terraform.tfstate files on exit
      disable_prevent_destroy: set all prevent destroy to false
      workspace_name: name of workspace to create or select

    Returns:
      Terraform init output.
    """
    # remove lifecycle prevent destroy
    if disable_prevent_destroy:
      min_python = (3, 5)
      if sys.version_info < min_python:
        raise TerraformTestError(
            'The disable_prevent_destroy flag requires at least Python 3.5')
      for tf_file in Path(self.tfdir).rglob('*.tf'):
        try:
          shutil.copy(str(tf_file), f'{str(tf_file)}.bkp')
        # except (OSError, IOError) as exc:
        except (OSError, IOError):
          _LOGGER.exception(
              f'Unable to backup terraform file {tf_file.resolve()}')
          raise TerraformTestError(
              f'Backup of terraform file ({tf_file.resolve()}) failed')
        try:
          with open(tf_file, 'r') as src:
            terraform = src.read()
          with open(tf_file, 'w') as src:
            terraform = re.sub(r'prevent_destroy\s+=\s+true',
                               'prevent_destroy = false', terraform)
            src.write(terraform)
        except (OSError, IOError):
          _LOGGER.exception(
              f'Unable to update prevent_destroy in file {tf_file.resolve()}')
          raise TerraformTestError(
              f'Unable to update prevent_destroy in file ({tf_file.resolve()}) failed'
          )

    # link extra files inside dir
    filenames = []
    for link_src in (extra_files or []):
      link_src = self._abspath(link_src)
      filename = os.path.basename(link_src)
      if os.path.isfile(link_src):
        link_dst = os.path.join(self.tfdir, filename)
        try:
          if os.name == 'nt':
            shutil.copy(link_src, link_dst)
          else:
            os.symlink(link_src, link_dst)
          filenames.append(filename)
        except FileExistsError as e:  # pylint:disable=undefined-variable
          _LOGGER.warning(e)
        else:
          _LOGGER.debug('linked %s', link_src)
      else:
        _LOGGER.warning('no such file {}'.format(link_src))
    self._finalizer = weakref.finalize(self, self._cleanup, self.tfdir,
                                       filenames, deep=cleanup_on_exit,
                                       restore_files=disable_prevent_destroy)
    setup_output = self.init(plugin_dir=plugin_dir, init_vars=init_vars,
                             backend=backend, use_cache=use_cache, **kw)
    if workspace_name:
      setup_output += self.workspace(name=workspace_name)
    return setup_output

  @_cache
  def init(self, input=False, color=False, force_copy=False, plugin_dir=None,
           init_vars=None, backend=True, use_cache=False, **kw):
    """Run Terraform init command."""
    cmd_args = parse_args(input=input, color=color, backend=backend,
                          force_copy=force_copy, plugin_dir=plugin_dir,
                          init_vars=init_vars, **kw)
    return self.execute_command('init', *cmd_args).out

  def workspace(self, name=None):
    """Run Terraform workspace command."""
    raw_ws_out = self.execute_command('workspace', *['list']).out
    cmd_args = ['select', name]
    if name not in [
        ws.replace('*', '').strip()
        for ws in raw_ws_out.split('\n')
        if len(ws) > 0
    ]:
      cmd_args = ['new', name]
    return self.execute_command('workspace', *cmd_args).out

  @_cache
  def plan(self, input=False, color=False, refresh=True, tf_vars=None,
           targets=None, output=False, tf_var_file=None, state=None,
           use_cache=False, **kw):
    """
    Run Terraform plan command, optionally returning parsed plan output.

    Args:
      input: Ask for input for variables if not directly set.
      no_color: If specified, output won't contain any color.
      refresh: Update state prior to checking for differences.
      tf_vars: Dict of variables in the Terraform configuration.
      targets: List of resources to target. Operation will be limited to this resource
        and its dependencies
      output: Determines if output will be returned.
      tf_var_file: Path to terraform variable configuration file relative to `self.tfdir`.
      state: Path to state file to use when reading the prior state snapshot.
    """
    cmd_args = parse_args(input=input, color=color, refresh=refresh,
                          tf_vars=tf_vars, targets=targets,
                          tf_var_file=tf_var_file, state=state, **kw)
    if not output:
      return self.execute_command('plan', *cmd_args).out
    with tempfile.NamedTemporaryFile() as fp:
      fp.close()
    # for tg we need to specify a temp name that is relative for the output to go into each
    # of the .terragrunt-cache, then plan / show would work, otherwise it overwrites each other!
    temp_file = fp.name if len(self._tg_ra()) == 0 else os.path.basename(
        fp.name)
    cmd_args.append('-out={}'.format(temp_file))
    self.execute_command('plan', *cmd_args)
    result = self.execute_command('show', '-no-color', '-json', temp_file)
    try:
      return self._plan_formatter(result.out)
    except json.JSONDecodeError as e:
      raise TerraformTestError(
          'Error decoding plan output: {}'.format(e))

  @_cache
  def apply(self, input=False, color=False, auto_approve=True, tf_vars=None,
            targets=None, tf_var_file=None, use_cache=False, **kw):
    """
    Run Terraform apply command.

    Args:
      input: Ask for input for variables if not directly set.
      no_color: If specified, output won't contain any color.
      auto_approve: Skip interactive approval of plan before applying.
      tf_vars: Dict of variables in the Terraform configuration.
      targets: List of resources to target. Operation will be limited to this resource
        and its dependencies
      tf_var_file: Path to terraform variable configuration file relative to `self.tfdir`.
    """
    cmd_args = parse_args(input=input, color=color, auto_approve=auto_approve,
                          tf_vars=tf_vars, targets=targets,
                          tf_var_file=tf_var_file, **kw)
    return self.execute_command('apply', *cmd_args).out

  @_cache
  def output(self, name=None, color=False, json_format=True, use_cache=False,
             **kw):
    """Run Terraform output command."""
    cmd_args = []
    if name:
      cmd_args.append(name)
    cmd_args += parse_args(color=color, json_format=json_format, **kw)
    output = self.execute_command('output', *cmd_args).out
    _LOGGER.debug('output %s', output)
    if json_format:
      try:
        output = self._output_formatter(output)
      except json.JSONDecodeError as e:
        _LOGGER.warning('error decoding output: {}'.format(e))
    return output

  @_cache
  def destroy(self, color=False, auto_approve=True, tf_vars=None, targets=None,
              tf_var_file=None, use_cache=False, **kw):
    """Run Terraform destroy command."""
    cmd_args = parse_args(color=color, auto_approve=auto_approve,
                          tf_vars=tf_vars, targets=targets,
                          tf_var_file=tf_var_file, **kw)
    return self.execute_command('destroy', *cmd_args).out

  def refresh(self, color=False, lock=False, tf_vars=None, targets=None, **kw):
    """Run Terraform refresh command."""
    cmd_args = parse_args(color=color, lock=lock, tf_vars=tf_vars,
                          targets=targets, **kw)
    return self.execute_command('refresh', *cmd_args).out

  def state_pull(self):
    """Pull state."""
    state = self.execute_command('state', 'pull')
    try:
      state = TerraformState(json.loads(state.out))
    except json.JSONDecodeError as e:
      _LOGGER.warning('error decoding state: {}'.format(e))
    return state

  def execute_command(self, cmd, *cmd_args):
    """Run arbitrary Terraform command."""
    _LOGGER.debug([cmd, cmd_args])
    cmdline = [self.binary, *self._tg_ra(), cmd]
    cmdline += cmd_args
    _LOGGER.info(cmdline)
    retcode = None
    full_output_lines = []
    try:
      stderr_mode = subprocess.STDOUT if os.name == 'nt' else subprocess.PIPE
      p = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=stderr_mode,
                           cwd=self.tfdir, env=self.env,
                           universal_newlines=True, encoding='utf-8',
                           errors='ignore')
      while True:
        output = p.stdout.readline()
        if output == '' and p.poll() is not None:
          break
        if output:
          _LOGGER.info(output.strip())
          full_output_lines.append(output)
      retcode = p.poll()
      p.wait()
    except FileNotFoundError as e:
      raise TerraformTestError('Terraform executable not found: %s' % e)
    out, err = p.communicate()
    full_output = "".join(full_output_lines)
    if retcode in [1, 11]:
      message = 'Error running command {command}: {retcode} {out} {err}'.format(
          command=cmd, retcode=retcode, out=full_output, err=err)
      _LOGGER.critical(message)
      raise TerraformTestError(message, err)
    return TerraformCommandOutput(retcode, full_output, err)

  def _tg_ra(self) -> List[str]:
    """if run_all return ['run-all'] else [] """
    return ['run-all'] if self._is_tg() and self.tg_run_all else []

  def _is_tg(self) -> bool:
    """based on the binary set determines if we are running terragrunt"""
    return self.binary.endswith('terragrunt')


def _parse_run_all_out(output: str, formatter: TerraformJSONBase) -> str:
  """
    run-all output a bunch of jsons back to back in one string(no comma),
    this convert the output to a valid json (put b2b jsons into a list)
  Args:
    output: the back to back jsons in a string
    formatter: output format, could be TerraformValueDict or TerraformPlanOutput
  Returns:
    convert the input into a list that is a valid json
  """
  dicts = json.loads("[" + re.sub(r"\}\s*\{", "}, {", output) + "]")
  return [formatter(d) for d in dicts]


class TerragruntTest(TerraformTest):

  def __init__(self, tfdir, basedir=None, binary='terragrunt', env=None,
               tg_run_all=False, enable_cache=False, cache_dir=None):
    """A helper class that could be used for testing terragrunt

    Most operations that apply to :func:`~TerraformTest` also apply to this class.
    Notice that to use this class for Terragrunt run-all, `tg_run_all` needs to be set to
    True.  The class would then only be used just for run-all.  If you need individual
    Terragrunt module testing, create another instance of this helper with
    tg_run_all=False (default)

    Args:
      tfdir: the Terraform module directory to test, either an absolute path, or
             relative to basedir.
      basedir: optional base directory to use for relative paths, defaults to the
               directory above the one this module lives in.
      binary: (Optional) path to terragrunt command.
      env: a dict with custom environment variables to pass to terraform.
      tg_run_all: whether the test is for terragrunt run-all, default to False
      enable_cache: Determines if the caching enabled for specific methods
      cache_dir: optional base directory to use for caching, defaults to
        the directory of the python file that instantiates this class
    """
    TerraformTest.__init__(self, tfdir, basedir, binary, env, enable_cache,
                           cache_dir)
    self.tg_run_all = tg_run_all
    if self.tg_run_all:
      self._plan_formatter = partial(_parse_run_all_out,
                                     formatter=TerraformPlanOutput)
      self._output_formatter = partial(_parse_run_all_out,
                                       formatter=TerraformValueDict)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test serialization of tftest instances"

import pytest
import tftest
import pickle


def assert_pickle_flow(obj):
  """Ensures instance is the same after being pickled"""
  pickled_obj = pickle.dumps(obj)
  pickled_obj = pickle.loads(pickled_obj)
  assert isinstance(pickled_obj, type(obj))


@pytest.fixture(scope="module")
def tf(fixtures_dir):
  tf = tftest.TerraformTest("plan_no_resource_changes", fixtures_dir)
  tf.setup()
  return tf


def test_setup(fixtures_dir):
  tf = tftest.TerraformTest("plan_no_resource_changes", fixtures_dir)
  expected = tf.setup()
  assert_pickle_flow(expected)


def test_plan(tf):
  expected = tf.plan(output=True)
  assert_pickle_flow(expected)


def test_apply(tf):
  expected = tf.apply()
  assert_pickle_flow(expected)


def test_output(tf):
  expected = tf.output()
  assert_pickle_flow(expected)


def test_state_pull(tf):
  expected = tf.state_pull()
  assert_pickle_flow(expected)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test the Terraform value dict wrapper class."

import pytest
import tftest


_RAW = {'a': {'value': 1, 'sensitive': True}, 'b': {'value': 2}}


@pytest.fixture
def wrapper():
  return tftest.TerraformValueDict(_RAW)


def test_getitem(wrapper):
  assert wrapper['a'] == 1


def test_getitem_dict_attrs(wrapper):
  assert wrapper.keys() == _RAW.keys()


def test_iter(wrapper):
  assert [k for k in wrapper] == [k for k in _RAW]



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test command out attribute of tftest exception class."

import pytest
import tftest


def test_cmd_err(fixtures_dir):
  with pytest.raises(tftest.TerraformTestError) as e:
    tf = tftest.TerraformTest('foobar')
    tf.setup()
  assert not e.value.cmd_error
  tf = tftest.TerraformTest('plan', fixtures_dir)
  tf.setup()
  with pytest.raises(tftest.TerraformTestError) as e:
    tf.plan()
  assert e.value.cmd_error



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test init and plan using an actual example."

import pytest
import tftest
import shutil
import os


@pytest.fixture(scope='module')
def terraform_test(fixtures_dir):
  tf = tftest.TerraformTest('plan_no_variables', fixtures_dir)
  tf.setup()
  yield tf
  shutil.rmtree(os.path.join(tf.tfdir, 'terraform.tfstate.d'))
  os.remove(os.path.join(tf.tfdir, '.terraform.lock.hcl'))


def test_new_workspace(terraform_test):
  tf_output = terraform_test.workspace(name="workspace_test")
  assert 'Created and switched to workspace "workspace_test"' in tf_output


def test_select_workspace(terraform_test):
  tf_output = terraform_test.workspace(name="workspace_test")
  tf_output = terraform_test.workspace(name="default")
  assert 'Switched to workspace "default"' in tf_output


def test_setup_with_workspace(terraform_test):
  tf_output = terraform_test.setup(workspace_name="setup_workspace")
  assert 'Created and switched to workspace "setup_workspace"' in tf_output



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test the function for mapping Terraform arguments."
import pytest

import tftest


ARGS_TESTS = (
    ({'auto_approve': True}, ['-auto-approve']),
    ({'auto_approve': False}, []),
    ({'backend': True}, []),
    ({'backend': None}, []),
    ({'backend': False}, ['-backend=false']),
    ({'color': True}, []),
    ({'color': False}, ['-no-color']),
    ({'color': False, 'input': False}, ['-no-color', '-input=false']),
    ({'force_copy': True}, ['-force-copy']),
    ({'force_copy': None}, []),
    ({'force_copy': False}, []),
    ({'input': True}, []),
    ({'input': False}, ['-input=false']),
    ({'json_format': True}, ['-json']),
    ({'json_format': False}, []),
    ({'lock': True}, []),
    ({'lock': False}, ['-lock=false']),
    ({'plugin_dir': ''}, []),
    ({'plugin_dir': 'abc'}, ['-plugin-dir', 'abc']),
    ({'refresh': True}, []),
    ({'refresh': None}, []),
    ({'refresh': False}, ['-refresh=false']),
    ({'upgrade': True}, ['-upgrade']),
    ({'upgrade': False}, []),
    ({'tf_var_file': None}, []),
    ({'tf_var_file': 'foo.tfvar'}, ['-var-file=foo.tfvar']),
    ({'tf_var_file': ['foo.tfvar', 'bar.tfvar']}, [
     '-var-file=foo.tfvar', '-var-file=bar.tfvar']),
    ({'tf_vars': {'text': 'text'}}, ['-var', 'text=text']),
    ({'tf_vars': {'number': 0}}, ['-var', 'number=0']),
    ({'tf_vars': {'bool': False}}, ['-var', 'bool=False']),
    ({'tf_vars': {'dict': {'text': 'text'}}},
     ['-var', 'dict={"text": "text"}']),
    ({'tf_vars': {'list': ['item1', 'item2']}},
     ['-var', 'list=["item1", "item2"]']),
    ({'tf_vars': {'dict': {'list': ['item1', 'item2']}}},
     ['-var', 'dict={"list": ["item1", "item2"]}']),
)


@pytest.mark.parametrize("kwargs, expected", ARGS_TESTS)
def test_args(kwargs, expected):
  assert tftest.parse_args() == []
  assert tftest.parse_args(**kwargs) == expected


TERRAGRUNT_ARGS_TESTCASES = [
    ({"tg_config": "Obama"}, ['--terragrunt-config', 'Obama']),
    ({"tg_tfpath": "Barrack"}, ['--terragrunt-tfpath', 'Barrack']),
    ({"tg_no_auto_init": True}, ['--terragrunt-no-auto-init']),
    ({"tg_no_auto_init": False}, []),
    ({"tg_no_auto_retry": True}, ['--terragrunt-no-auto-retry']),
    ({"tg_no_auto_retry": False}, []),
    ({"tg_non_interactive": True}, ['--terragrunt-non-interactive']),
    ({"tg_non_interactive": False}, []),
    ({"tg_working_dir": "George"}, ['--terragrunt-working-dir', 'George']),
    ({"tg_download_dir": "Bush"}, ['--terragrunt-download-dir', 'Bush']),
    ({"tg_source": "Clinton"}, ['--terragrunt-source', 'Clinton']),
    ({"tg_source_update": True}, ['--terragrunt-source-update']),
    ({"tg_source_update": False}, []),
    ({"tg_iam_role": "Bill"}, ['--terragrunt-iam-role', 'Bill']),
    ({"tg_ignore_dependency_errors": True}, [
     '--terragrunt-ignore-dependency-errors']),
    ({"tg_ignore_dependency_errors": False}, []),
    ({"tg_ignore_dependency_order": True}, [
     '--terragrunt-ignore-dependency-order']),
    ({"tg_ignore_dependency_order": False}, []),
    ({"tg_ignore_external_dependencies": "dont care what is here"},
     ['--terragrunt-ignore-external-dependencies']),
    ({"tg_include_external_dependencies": True}, [
     '--terragrunt-include-external-dependencies']),
    ({"tg_include_external_dependencies": False}, []),
    ({"tg_parallelism": 20}, ['--terragrunt-parallelism 20']),
    ({"tg_exclude_dir": "Ronald"}, ['--terragrunt-exclude-dir', 'Ronald']),
    ({"tg_include_dir": "Reagan"}, ['--terragrunt-include-dir', 'Reagan']),
    ({"tg_check": True}, ['--terragrunt-check']),
    ({"tg_check": False}, []),
    ({"tg_hclfmt_file": "Biden"}, ['--terragrunt-hclfmt-file', 'Biden']),
    ({"tg_override_attr": {"Iron": "Man", "Captain": "America"}},
     ['--terragrunt-override-attr=Iron=Man', '--terragrunt-override-attr=Captain=America']),
    ({"tg_debug": True}, ['--terragrunt-debug']),
    ({"tg_debug": False}, []),

]


@pytest.mark.parametrize("kwargs, expected", TERRAGRUNT_ARGS_TESTCASES)
def test_terragrunt_args(kwargs, expected):
  assert tftest.parse_args(**kwargs) == expected


def test_var_args():
  assert sorted(tftest.parse_args(init_vars={'a': 1, 'b': '["2"]'})) == sorted(
      ["-backend-config=a=1", '-backend-config=b=["2"]'])
  assert sorted(tftest.parse_args(tf_vars={'a': 1, 'b': '["2"]'})) == sorted(
      ['-var', 'b=["2"]', '-var', 'a=1'])


def test_targets():
  assert tftest.parse_args(targets=['one', 'two']) == sorted(
      ['-target=one', '-target=two'])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test the plan output wrapper classes."

import pytest
import tftest


@pytest.fixture
def run_init(fixtures_dir):

  def init_runner(**backend_config):
    tf = tftest.TerraformTest('apply', fixtures_dir)
    tf.setup(init_vars=backend_config)

  return init_runner


def test_simple_args(run_init):
  run_init(path='terraform-1.tfstate')


def test_spaced_args(run_init):
  run_init(path='terraform 1.tfstate')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test the Terraform state wrapper class."

import pytest
import tftest


@pytest.fixture
def state(fixtures_dir):
  import json
  with open('%s/state.json' % fixtures_dir) as fp:
    return tftest.TerraformState(json.load(fp))


def test_attributes(state):
  assert state.version == 4
  assert state.terraform_version == '0.12.10'


def test_outputs(state):
  assert state.outputs['foo'] == 'foo-value'


def test_resources(state):
  res = state.resources['module.vpc-remote.google_compute_network.network']
  assert res['mode'] == 'managed'
  assert res['instances'][0]['attributes']['name'] == 'remote'



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test init and plan with prevent_destroy lifecycles"
import logging
from unittest.mock import mock_open, patch

import pytest
import tftest


def test_with_no_lifecycle_override(fixtures_dir):
  tf = tftest.TerraformTest('prevent_destroy', fixtures_dir)
  tf.setup()
  tf.apply(auto_approve=True)
  with pytest.raises(tftest.TerraformTestError):
    tf.destroy(auto_approve=True)
    tf.output()


def test_with_with_lifecycle_override(fixtures_dir):
  tf = tftest.TerraformTest('prevent_destroy', fixtures_dir)
  tf.setup(disable_prevent_destroy=True)
  tf.apply(auto_approve=True)
  tf.destroy(auto_approve=True)


@patch('tftest.shutil.copy')
def test_raises_exception_if_file_backup_fails(mock_copy, fixtures_dir, caplog):
  caplog.set_level(logging.ERROR)
  tf = tftest.TerraformTest('prevent_destroy', fixtures_dir)
  mock_copy.side_effect = [FileNotFoundError(
      '[Errno 2] No such file or directory: "sddsdsds"')]
  with pytest.raises(tftest.TerraformTestError):
    tf.setup(disable_prevent_destroy=True)
  assert caplog.messages[0].startswith('Unable to backup terraform file ')


@patch('tftest.shutil.copy')
def test_raises_exception_when_read_or_write_fails(mock_copy, fixtures_dir, caplog):
  caplog.set_level(logging.ERROR)
  tf = tftest.TerraformTest('prevent_destroy', fixtures_dir)
  with patch('builtins.open', mock_open(read_data='')) as mock_file:
    mock_file.return_value.write.side_effect = [
        IOError('[Errno 13] Permission denied: "sddsdsds"')]
    with pytest.raises(tftest.TerraformTestError):
      tf.setup(disable_prevent_destroy=True)
  assert caplog.messages[0].startswith(
      'Unable to update prevent_destroy in file ')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools

from tftest import __version__


with open("README.md", "r") as fh:
  long_description = fh.read()


setuptools.setup(
    name="tftest",
    version=__version__,
    author="Ludovico Magnocavallo",
    author_email="ludomagno@google.com",
    description="Simple Terraform test helper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GoogleCloudPlatform/terraform-python-testing-helper",
    py_modules=['tftest'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    setup_requires=['nose>=1.3'],
    test_suite='nose.collector',
)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test init and plan using an actual example."

import pytest
import tftest


@pytest.fixture(scope="module")
def plan(fixtures_dir):
  tf = tftest.TerraformTest('plan', fixtures_dir)
  tf.setup(extra_files=['plan.auto.tfvars'])
  return tf.plan(output=True)


def test_variables(plan):
  assert 'prefix' in plan.variables
  assert plan.variables['name'] == 'test'


def test_outputs(plan):
  assert 'service_account' in plan.outputs


def test_root_resource(plan):
  res = plan.resources['google_project_iam_member.test_root_resource']
  assert res['values']['project'] == plan.variables['project_id']


def test_modules(plan):
  mod = plan.modules['module.service-account']
  assert set(mod.resources.keys()) == set(
      ['google_service_account.service_account'])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Test the plan output wrapper classes."

import pytest
import tftest


@pytest.fixture(scope="module")
def plan_out(fixtures_dir):
  import json
  with open('%s/plan_output.json' % fixtures_dir) as fp:
    return tftest.TerraformPlanOutput(json.load(fp))


@pytest.fixture(scope="module")
def plan_out_no_prior(fixtures_dir):
  import json
  with open('%s/plan_output_no_prior.json' % fixtures_dir) as fp:
    return tftest.TerraformPlanOutput(json.load(fp))


def test_output_attributes(plan_out):
  assert plan_out.format_version == "0.1"
  assert plan_out.terraform_version == "0.12.6"


def test_variables(plan_out):
  assert plan_out.variables['foo'] == 'bar'


def test_resource_changes(plan_out):
  address = 'module.resource-change.foo_resource.somefoo'
  change = plan_out.resource_changes[address]
  assert change['address'] == address
  assert change['change']['before'] is None


def test_output_changes(plan_out):
  change = plan_out.output_changes['spam']
  assert change['after'] == 'bar'


def test_configuration(plan_out):
  assert plan_out.configuration['provider_config']['google']['name'] == 'google'


def test_root_module(plan_out):
  mod = plan_out.root_module
  assert plan_out.modules == mod.child_modules
  assert plan_out.resources == mod.resources


def test_resources(plan_out):
  res = plan_out.resources['spam.somespam']
  assert res['address'] == 'spam.somespam'
  assert res['values']['spam-value'] == 'spam'


def test_modules(plan_out):
  mod = plan_out.modules['module.parent']
  assert mod['address'] == 'module.parent'
  res = mod.resources['foo.somefoo']
  assert res['address'] == 'module.parent.foo.somefoo'
  assert res['values']['foo-value'] == 'foo'


def test_child_modules(plan_out):
  mod = plan_out.modules['module.parent'].child_modules['module.child']
  assert mod.resources['eggs.someeggs']['values']['eggs-value'] == 'eggs'


def test_plan_with_no_resources_succeeds(fixtures_dir):
  tf = tftest.TerraformTest('plan_no_resource_changes', fixtures_dir)
  result = tf.plan(output=True)

  assert result.outputs['just_an_output'] == 'Hello, plan!'


def test_plan_stdout(fixtures_dir):
  tf = tftest.TerraformTest('plan_no_resource_changes', fixtures_dir)
  result = tf.plan(output=False)
  assert 'just_an_output = "Hello, plan!"' in result


def test_plan_prior_state(plan_out):
  prior_resources = plan_out.prior_resources
  assert len(prior_resources) == 1
  assert prior_resources['data.google_client_config.current']['type'] == 'google_client_config'
  assert prior_resources['data.google_client_config.current']['values']['description'] == 'foo-value'


def test_plan_no_prior_state(plan_out_no_prior):
  prior_resources = plan_out_no_prior.prior_resources
  assert not prior_resources
